@@ -26,24 +26,16 @@ my %render = (
     combo   => sub { return Wx::GridCellChoiceEditor->new(@_); },
     boolean => sub { return Wx::GridCellBoolRenderer->new; },
     time    => sub { return Wx::GridCellDateTimeRenderer->new; },
+    
 );
 
-my %fieldtype = (
-    serial  => "number",
-    varchar => "text",
-    char    => "text",
-    integer => "number",
-    enum    => "text",
-    date    => "time",
-    boolean => "boolean",
-    set     => "text",
-);
 use constant {
     UNCHANGED => 0,
     CHANGED   => 1,
     INSERTED  => 2,
     DELETED   => 3,
     LOCKED    => 4,
+    STATUS_COLUMN => -1,
 };
 
 #use constant STATUS_LAB => qw(sync !!! new del lck);
@@ -77,13 +69,27 @@ sub new {
     $self->{hcols} = \%hcols;
     $self->{log}->debug( "new called - cols: " . join( " ", @cols ) );
 
-    $self->_setup_fields;
-    my $table = $self->_setup_gridtable;
+    # $self->_setup_fields;
+      
 
     #my $table =
     #$table->SetView($self);
-    $self->SetTable($table);
+    #$self->SetTable($table);
 
+    $self->{ds_helper} = Gtk2::Ex::DbLinker::DatasheetHelper->new(
+        cols       => $self->{cols},
+        dman       => $self->{dman},
+	#col_number => sub { $self->colnumber_from_name( $_[0] ); }
+
+    );
+  ( $self->{fields}, $self->{hiddencols}) = 
+       $self->{ds_helper}->setup_fields(
+	    allfields=> $self->{fields}, 
+	    cols=> $self->{cols}
+            ) ;
+	    
+    my $table = $self->_setup_gridtable; 
+    $self->SetTable($table);
     $self->_setup_grid;
     $self->SetSelectionMode(wxGridSelectRows);
 
@@ -91,15 +97,8 @@ sub new {
     EVT_GRID_SELECT_CELL( $self, sub { shift->_cell_clicked(@_); } );
     EVT_GRID_RANGE_SELECT( $self, sub { shift->_row_selected(@_); } );
     $self->{log}->debug( "constructor done: rows: " . $self->GetNumberRows );
-	
-    $self->{ds_helper} = Gtk2::Ex::DbLinker::DatasheetHelper->new(
-        cols       => $self->{cols},
-        dman       => $self->{dman},
-	col_number => sub { $self->colnumber_from_name($_[0]);}
-	
 
-    );
-#col_number => \&{ $self->colnumber_from_name }
+    #col_number => \&{ $self->colnumber_from_name }
     $self->{changed} = 0;
     return $self;
 
@@ -111,59 +110,6 @@ sub _setup_gridtable {
 
 }
 
-sub _setup_fields {
-    my ($self) = @_;
-
-    #die Dumper(@{$self->{fields}});
-    my @hiddencols;
-    if ( !$self->{fields} ) {
-        my $no_of_fields = scalar @{ $self->{cols} };
-        my $field_percentage =
-          $no_of_fields < 8
-          ? 100 / $no_of_fields
-          : 12.5;    # Don't set percentages < 12.5 - doesn't really work so well ...
-        for my $field ( @{ $self->{cols} } ) {
-            my $gtktype = $fieldtype{ $self->{dman}->get_field_type($field) };
-            push @{ $self->{fields} },
-              {
-                name      => $field,
-                x_percent => $field_percentage,
-                renderer  => $gtktype,
-              };
-            $self->{log}->debug( " * set field : " . $field . " renderer : " . $gtktype );
-        }
-    }
-
-    my $column_no = 0;
-    for my $field ( @{ $self->{fields} } ) {
-        $self->{log}->debug( "field name : " . $field->{name}, " col_number: ", $column_no );
-
-        $self->{colname_to_number}->{ $field->{name} } = $column_no;
-        if ( defined $field->{renderer} && $field->{renderer} eq "hidden" ) {
-            push @hiddencols, $column_no;
-        }
-
-        if ( !$field->{renderer} ) {
-
-            #my $x = ( $fieldtype{ $self->{fieldsType}->{$field->{name}}}  ?  $fieldtype{$self->{fieldsType}->{$field->{name}}} : "text");
-            my $ftype = $self->{dman}->get_field_type( $field->{name} );
-            my $x = ( $fieldtype{$ftype} ? $fieldtype{$ftype} : "text" );
-            $field->{renderer} = $x;
-            $self->{log}->debug( "reset renderer for field " . $field->{name} . " to " . $x );
-
-        }
-
-        #if ($field->{renderer} eq "combo") {
-        #    $self->setup_combo($field->{name});
-        #}
-        $self->{log}->debug( " ** set field : " . $field->{name} . " renderer : " . $field->{renderer} );
-        $field->{column} = $column_no++;
-    }
-
-    #setupfields_end
-    $self->{hiddencols} = \@hiddencols;
-}    # setup_fields
-
 #combo are backed up with two array ref and a scalar:
 #$self->{combo_val}->{ $field_name } : array ref of values diplayed
 #self->{combo_id}->{ $field_name } : array ref of corresponding id (that are stored in the table in the $field_name column
@@ -180,24 +126,11 @@ sub _setup_grid {
         #my $renderer = $render{$field->{renderer}}[0]();
         my $ed;
 
-        #	my $cell_ref = ref $renderer;
-
-        #if   ($field->{renderer} eq "hidden") {
-        #my $colnum = $self->{colname_to_number}->{ $field->{name} };
-        #$self->SetColSize($colnum, 0);
-        #} els
-
         if ( $field->{renderer} eq "toggle" ) {
             $field->{editor} = $render{ $field->{renderer} }();
         } elsif ( $field->{renderer} eq "combo" ) {
 
             $field->{editor} = $render{ $field->{renderer} }( $self->_setup_combo( $field->{name} ), 0 );
-
-            #$field->{editor} = Wx::GridCellChoiceEditor->new(  $self->_setup_combo($field->{name}),0 );
-
-            #$self->SetCellRenderer($row, $col, $ed);
-            #$renderer->set(  editable => 1, text_column     => 1,	has_entry   => 0,  model => $model );
-            # $renderer->{col_data} = $lastcol++ ;
             $self->{log}->debug( "field name with combo renderer: " . $field->{name} );
 
             my $signal = $signals{ ref $field->{editor} };
@@ -227,7 +160,7 @@ sub _setup_grid {
 
         } else {
             $self->{log}->debug( "field name with txt renderer: " . $field->{name} );
-            $field->{editor} = $render{ $field->{renderer} }();
+            $field->{editor} = $render{ $field->{renderer} }() if ($render{ $field->{renderer} });
 
             if ( grep /^$field->{name}$/, @apk ) {
                 $self->{log}->debug("not editable because it's a pk");
@@ -248,7 +181,8 @@ sub _setup_grid {
 
         #$grid->SetReadOnly($r, 0);
         for my $field ( @{ $self->{fields} } ) {
-            my $col = $self->{colname_to_number}->{ $field->{name} };
+		#my $col = $self->{colname_to_number}->{ $field->{name} };
+		my $col = $self->{ds_helper}->colnumber_from_name( $field->{name} );
             if ( $r == 0 ) {
 
                 my $label = (
@@ -258,6 +192,7 @@ sub _setup_grid {
                 );
                 $self->{log}->debug( "col: ", $col, " label: ", $label );
                 $self->{log}->debug( "cell editor: " . ref $field->{editor} );
+		die unless defined $col;
                 $self->SetColLabelValue( $col, $label );
                 if ( defined $field->{size} ) {
                     $self->SetColSize( $col, $field->{size} );    #if ($r==0);
@@ -327,74 +262,19 @@ sub _setup_grid {
 
 }    #setup_grid
 
-#the first field links the data from the table with a value in the list
-#the remaining fields are displayed in the combo
-# the type of the first field is in ->{fieldsType} the other(s) are supposed to be strings
 sub _setup_combo {
     my ( $self, $fieldname ) = @_;
+     my $column_no =  $self->colnumber_from_name($fieldname);
+    my $fields = $self->{fields}[$column_no ];
+    my $col_ref = $self->{ds_helper}->init_combo_setup(name=> $fieldname, fields => $fields);	
+    my ( $rowsref, $idsref ) = $self->{ds_helper}->setup_combo( fields => $fields, name => $fieldname, col_ref=>$col_ref);
 
-    my $column_no = $self->{colname_to_number}->{$fieldname};
-    $self->{log}->debug( "setup_combo field name : " . $fieldname . " col number :" . $column_no );
+    $self->{combo_id}->{$fieldname}  = $idsref;
+    $self->{combo_val}->{$fieldname} = $rowsref;
+    $self->{combo_cur}               = [];
 
-    #my @combo = @{$self->{fields}[$column_no]->{data}};
-    my $dman = $self->{fields}[$column_no]->{data_manager};
-    my $last = $dman->row_count;
-
-    # return unless(@combo);
-
-    my @liste_def;
-
-    # my $firstrow = $combo[0];
-
-    my @cols;    # = $firstrow->meta->column_names;
-    if ( $self->{fields}[$column_no]->{fieldnames} ) {
-        @cols = @{ $self->{fields}[$column_no]->{fieldnames} };
-    } else {
-        @cols = $dman->get_field_names;
-    }
-
-    $self->{log}->debug( "setup_combo cols : " . join( " ", @cols ) );
-
-    my @rows;
-    my @ids;
-    for ( my $row_pos = 0 ; $row_pos < $last ; $row_pos++ ) {
-
-        #$self->{log}->debug("Datasheet setupcombo set row pos " . $row_pos);
-        $dman->set_row_pos($row_pos);
-
-        #push @model_row, $model->append;
-        my $lastcol = @cols - 1;
-        my $rowval;
-        my $pos = 0;
-        foreach my $name (@cols) {
-
-            #push @model_row, $pos++, $row->$name();
-            if ( $pos == 0 ) { push @ids, $dman->get_field($name); }
-            else {
-                $rowval .= $dman->get_field($name) . " ";
-            }
-            push @rows, $rowval if ( $pos == $lastcol );
-            $pos++;
-
-            # $self->{log}->debug("field: " . $name . " val : ". $dman->get_field($name));
-        }
-
-        # print Dumper(@model_row);
-        #$model->set( @model_row );
-        # $model->set_text_column(1);
-
-    }    #for
-
-    #  fieldname est le nom du champ li au combo dans la table affichee par la grid
-    #  il est unique pour la table contrairement aux champs qui sont dans la combo
-    $self->{combo_id}->{$fieldname}  = \@ids;
-    $self->{combo_val}->{$fieldname} = \@rows;
-
-    #$self->{combo_cur}->{$fieldname} = [];
-    $self->{combo_cur} = [];
-    return \@rows;
-
-}    #setup_combo
+    return $rowsref;
+}
 
 sub get_column_value {
 
@@ -469,8 +349,9 @@ sub get_current_row {
 sub colnumber_from_name {
 
     my ( $self, $fieldname ) = @_;
-    confess "fieldname undef" unless ( defined $fieldname);
-    return $self->{colname_to_number}->{$fieldname}
+    #confess "fieldname undef" unless ( defined $fieldname );
+    #return $self->{colname_to_number}->{$fieldname}
+    return $self->{ds_helper}->colnumber_from_name($fieldname);
 
 }
 
@@ -525,18 +406,18 @@ sub apply {
     my $last  = $self->GetNumberRows;
 
     #$self->{ds_helper}->init_rows($self);
-    $self->{ds_helper}->init(
+    $self->{ds_helper}->init_apply(
         iter         => 0,
         next         => sub { my $p = $_[0]; $p++; return ( $p < $last ? $p : undef ); },
         has_more_row => sub { my $v = ( $self->{iter} < $last ); $self->{log}->debug( "has_more_row ", ( $v ? " true" : " false" ) ); return $v; },
         get_val      => sub {
             my ( $row, $col, $name ) = @_;
             my $val;
-            if ( $col == 0 ) {
+	    if ( $col == STATUS_COLUMN ) {
                 my $status_lab = $self->GetRowLabelValue($row);
 
                 $val = $self->_find_index( $status_lab, (STATUS_LAB) );
-            } else {
+	     } else {
                 if ( $self->_is_combo($col) ) {
                     my $aref = $self->{combo_cur}->{$name};
                     $val = $aref->[$row];
@@ -544,14 +425,14 @@ sub apply {
                 } else {
                     $val = $self->GetTable->GetValue( $row, $col );
                 }
-            }
+	     }
             $self->{log}->debug( "get_val found: ", ( defined $val ? $val : "undef" ), " for row : ", $row, " col : ", $col );
             return $val;
         },
         set_val => sub {
             my ( $row, $col, $val ) = @_;
             $self->{log}->debug( "set_val : ", ( defined $val ? $val : " undef" ), " for row : ", $row, " col : ", $col );
-            if ( $col == 0 ) {
+            if ( $col == STATUS_COLUMN ) {
                 my $label = (STATUS_LAB)[$val];
                 $self->{log}->debug( "label found : ", $label );
                 $self->SetRowLabelValue( $row, $label );
@@ -562,7 +443,8 @@ sub apply {
         del_row => sub {
             $self->{log}->debug( "deleting at row ", $_[0] );
             $self->DeleteRows( $_[0] );
-        }
+        },
+	status_col => STATUS_COLUMN,
     );
 
     #$self->{ds_helper}->{dman} = $self->{dman};
@@ -572,178 +454,13 @@ sub apply {
 
 }
 
-=for comment
-sub apply_old {
-
-    my $self = shift;
-
-    #my $pkref = ( defined $_[0] ? $_[0] : undef );
-    my $pkref = $_[0];
-
-    #accder au lignes modifiees: parcourir toutes celles qui sont affiches
-    #et agir selon ce qu'indique la col status
-
-    $self->{changed} = 0;
-    $self->{log}->debug("apply");
-    my @iters_to_remove;
-
-    #my $model = $self->{treeview}->get_model;
-
-    #my $iter = $model->get_iter_first;
-    #my $row_pos = 0;
-    my $last = $self->GetNumberRows - 1;
-    my $row;
-
-    my @fields_to_save;
-    if ($pkref) {
-        my %seen;    #remove from cols the fields received in arg
-        %seen = map { $_ => $seen{$_}++ } ( @{ $self->{cols} }, @{$pkref} );
-        @fields_to_save = grep { $seen{$_} < 1 } keys %seen;
-    } else {
-        @fields_to_save = @{ $self->{cols} };
-    }
-
-    # $model->signal_handler_block( $self->{changed_signal} );
-    #for all the rows in the datasheet
-
-    for my $row_pos ( 0 .. $last ) {
-        my $status_lab = $self->GetRowLabelValue($row_pos);
-
-        #my $status = $model->get( $iter, STATUS_COLUMN );
-        my $status = $self->_find_index( $status_lab, (STATUS_LAB) );
-
-        $self->{log}->debug( "status_lab: "
-              . $status_lab
-              . " status : "
-              . $status
-              . " row pos "
-              . $row_pos );
-
-        $self->{dman}->set_row_pos($row_pos);
-
-        # Decide what to do based on status
-        if ( $status == UNCHANGED || $status == LOCKED ) {
-
-            #$iter = $model->iter_next( $iter );
-            #$row_pos++;
-            next;
-        }
-
-        if ( $status == INSERTED ) {    # new row for the database
-            $self->{dman}->new_row;
-
-        }    #else { # existing row with CHANGED or DELETED
-             # $row = ${$self->{data}}[$row_pos++];
-             #}
-
-        if ( $status == DELETED ) {
-
-#if ($row->delete ) {$self->{log}->debug("deleting current row");} else {$self->{log}->debug("Can't delete");}
-#if ( $self->{dman}->delete ) {$self->{log}->debug("deleting current row");} else {$self->{log}->debug("Can't delete");}
-
-            $self->{log}->debug(
-" should delete row at pos $row_pos in the datamanager ... just testing for now"
-            );
-
-            push @iters_to_remove, $row_pos;
-        } else {    # changed, inserted
-            my $count_update;
-
-            #for my $field ( @{$self->{fields}} ) {
-            for my $name (@fields_to_save) {
-
-                #my $name = $field->{name};
-                if ( defined $self->{hcols}->{$name} ) {
-
- #my $x =  $model->get( $iter, $self->{colname_to_number}->{ $field->{name} } );
-                    my $col = $self->{colname_to_number}->{$name};
-                    my $x;
-                    if ( $self->_is_combo($col) ) {
-			#my $aref = $self->{combo_cur}->{$name};
-			my $aref = $self->{combo_cur};
-                        $x = $aref->[$row_pos];
-                        $self->{log}->debug( " found "
-                              . $x
-                              . " for grid row "
-                              . $row_pos
-                              . " for combo "
-                              . $name );
-                    } else {
-
-                        $x = $self->GetTable->GetValue( $row_pos, $col );
-                    }
-                    $self->{log}->debug( "Field: "
-                          . $name
-                          . " row_pos "
-                          . $row_pos
-                          . " value: "
-                          . ( defined $x ? $x : "undef" ) );
-                    $count_update++;
-                    $self->{dman}->set_field( $name, $x );
-
-                } else {
-                    $self->{log}->debug( "apply : "
-                          . $name
-                          . " not found in "
-                          . join( " ", @{ $self->{cols} } ) );
-                }
-
-            }
-            if ($count_update) {
-                $self->{log}->debug("saving...");
-
-                #$row->save;
-                $self->{dman}->save;
-            } else {
-                $self->{log}->debug("no update, not saving");
-            }
-        }
-
-        #replace the unchanged icon in the col 0
-        #$model->set( $iter, STATUS_COLUMN, UNCHANGED );
-        $self->SetRowLabelValue( $row_pos, (STATUS_LAB)[UNCHANGED] );
-
-        #$row_pos++;
-        #$iter = $model->iter_next( $iter );
-
-    }    #while
-
-    foreach my $pos (@iters_to_remove) {
-
-        #$model->remove( $iter );
-        $self->DeleteRows($pos);
-        $self->{dman}->set_row_pos($pos);
-        $self->{dman}->delete;
-    }
-
-    # $model->signal_handler_unblock( $self->{changed_signal} );
-    if ( defined $self->{after_update} ) {
-
-        $self->{after_update}();
-    }
-
-}    #apply
-=cut
-
 sub insert {
 
     my ( $self, @columns_and_values ) = @_;
     $self->{log}->debug("insert");
 
-    #return;
-    # my $model = $self->{treeview}->get_model;
-    #  my $iter = $model->append;
-
-    # print Dumper(@columns_and_values);
-    #my @new_record;
-
-    #push @new_record,        $iter,        STATUS_COLUMN,        INSERTED;
-
     $self->{log}->debug( "new rec default values: " . join( " ", @columns_and_values ) );
 
-    #$model->set( @new_record );
-
-    #$self->{treeview}->set_cursor( $model->get_path($iter), $self->{fields}[0]->{treeview_column}, 0 );
     $self->AppendRows(1);
     my $last = $self->GetNumberRows - 1;
 
@@ -755,14 +472,8 @@ sub insert {
         foreach my $col_no_or_value (@columns_and_values) {
 
             if ($i) {
-
-                # die (Dumper $self->{colname_to_number});
-                #die ($col_name_or_value);
-                #$col_no = $self->colnumber_from_name($col_name_or_value);
                 $col_no = $col_no_or_value;
-
-                #die ($col_no);
-                $i = 0;
+                $i      = 0;
             } else {
                 $self->{log}->debug( "last: " . $last . " col: " . $col_no . " value: " . $col_no_or_value );
                 $self->SetCellValue( $last, $col_no, $col_no_or_value );
@@ -838,11 +549,8 @@ sub _combo_edited {
     my @a = @{$aref};
     $self->{log}->debug( "id: " . Dumper($id) );
     $self->{log}->debug( "new value:  " . $new_text . " index: " . $i . " corresponding id : " . $a[$i] );
-
     #$self->{combo_cur}->{$id } = $a[$i];
-
     #update the array backing the combo column in the grid
-
     #$self->{combo_cur}->{$id}->[$current_row] = $a[$i];
     $self->{combo_cur}->[$current_row] = $a[$i];
 
@@ -908,7 +616,7 @@ sub new {
     $self->{dman}    = $args->{dman};
     $self->{fields}  = $args->{fields};
     $self->{grid}    = $args->{grid};
-    $self->{cols}    = scalar @{ $self->{fields} };
+    $self->{cols}    = scalar @{ $self->{fields} };  #here cols holds the number of fields ... in Wxdatasheet it's an ar of the fields names
     $self->{coldata} = [];
     $self->{array}   = [];
     $self->{rows}    = $self->{dman}->row_count();
